@@ -12,20 +12,28 @@
 #include <dirent.h>
 
 struct _MicrofeedSubscriber {
+	char* identifier;
 	DBusConnection* connection;
 	char* object_path;
 	MicrofeedConfiguration* configuration;
 	
 	MicrofeedStore* publishers_by_identifier;
-	MicrofeedStore* publishers_by_unique_connection_name;
+	MicrofeedStore* providers_by_bus_name;
+	MicrofeedStore* providers_by_unique_connection_name;
 };
 
 typedef struct {
 	MicrofeedSubscriber* subscriber;
-	char* identifier;
-	char* object_path;
 	char* bus_name;
 	char* unique_connection_name;
+	MicrofeedStore* publishers_by_object_path;
+} Provider;
+
+typedef struct {
+	unsigned int reference_count;
+	Provider* provider;
+	char* identifier;
+	char* object_path;
 	time_t last_activity;
 	MicrofeedStore* feeds;
 } Publisher;
@@ -52,10 +60,22 @@ typedef struct {
 	void* user_data;
 } MethodReturnData;
 
-static Publisher* publisher_new(MicrofeedSubscriber* subscriber, const char* identifier);
+static Provider* provider_new(MicrofeedSubscriber* subscriber, const char* bus_name);
+static void provider_free(Provider* provider);
+static const char* provider_get_bus_name(Provider* provider);
+static const char* provider_get_unique_connection_name(Provider* provider);
+static void provider_set_unique_connection_name(Provider* provider, const char* unique_connection_name);
+
+static Publisher* publisher_new(Provider* provider, const char* identifier);
 static void publisher_free(Publisher* publisher);
+Publisher* publisher_ref(Publisher* publisher);
+void publisher_unref(Publisher* publisher);
+static const char* publisher_get_identifier(Publisher* publisher);
+static const char* publisher_get_object_path(Publisher* publisher);
+
 static Feed* feed_new(Publisher* publisher, const char* uri, MicrofeedSubscriberCallbacks* callbacks, void* user_data);
 static void feed_free(Feed* feed);
+static const char* feed_get_uri(Feed* feed);
 
 static void call_publisher_method(Publisher* publisher, const char* uri, const char* uid, DBusMessage* message, MicrofeedSubscriberErrorCallback callback, void* user_data);
 static MicrofeedItem* parse_item_from_message(DBusMessageIter* iter);
@@ -63,9 +83,6 @@ static int parse_item_signal(Publisher* publisher, DBusMessage* message, Feed**
 static int parse_feed_signal(Publisher* publisher, DBusMessage* message, Feed** feed_return);
 static Publisher* get_publisher(MicrofeedSubscriber* subscriber, const char* publisher_identifier, const char* uri, const char* uid, MicrofeedSubscriberErrorCallback callback, void* user_data);
 static Feed* get_feed(MicrofeedSubscriber* subscriber, const char* publisher_identifier, const char* uri, const char* uid, MicrofeedSubscriberErrorCallback callback, void* user_data);
-static const char* feed_get_uri(Feed* feed);
-static const char* publisher_get_identifier(Publisher* publisher);
-static const char* publisher_get_unique_connection_name(Publisher* publisher);
 static void object_unregister(DBusConnection* connection, void* user_data);
 static DBusHandlerResult object_message(DBusConnection* connection, DBusMessage* message, void* user_data);
 
@@ -99,18 +116,21 @@ static DBusObjectPathVTable object_vtable = {
 	object_message
 };	
 
-MicrofeedSubscriber* microfeed_subscriber_new(const char* object_path, MicrofeedConfiguration* configuration, DBusConnection* connection) {
+MicrofeedSubscriber* microfeed_subscriber_new(const char* identifier, const char* object_path, MicrofeedConfiguration* configuration, DBusConnection* connection) {
 	MicrofeedSubscriber* subscriber;
 	
 	subscriber = microfeed_memory_allocate(MicrofeedSubscriber);
+	subscriber->identifier = strdup(identifier);
 	subscriber->connection = connection;
 	subscriber->object_path = strdup(object_path);
 	subscriber->configuration = configuration;
 	
 	subscriber->publishers_by_identifier = microfeed_store_new_sorted((MicrofeedStoreCompareKeysFunction)strcmp,
-	                                                           (MicrofeedStoreGetKeyFunction)publisher_get_identifier);
-	subscriber->publishers_by_unique_connection_name = microfeed_store_new_sorted((MicrofeedStoreCompareKeysFunction)strcmp,
-	                                                                       (MicrofeedStoreGetKeyFunction)publisher_get_unique_connection_name);
+	                                                                  (MicrofeedStoreGetKeyFunction)publisher_get_identifier);
+	subscriber->providers_by_bus_name = microfeed_store_new_sorted((MicrofeedStoreCompareKeysFunction)strcmp,
+	                                                               (MicrofeedStoreGetKeyFunction)provider_get_bus_name);
+	subscriber->providers_by_unique_connection_name = microfeed_store_new_sorted((MicrofeedStoreCompareKeysFunction)strcmp,
+	                                                                             (MicrofeedStoreGetKeyFunction)provider_get_unique_connection_name);
 
 	if (!dbus_connection_register_object_path(subscriber->connection, subscriber->object_path, &object_vtable, subscriber)) {
 	
@@ -128,6 +148,11 @@ void microfeed_subscriber_free(MicrofeedSubscriber* subscriber) {
 	/* TODO */
 }
 
+const char* microfeed_subscriber_get_identifier(MicrofeedSubscriber* subscriber) {
+
+	return subscriber->identifier;
+}
+
 int microfeed_subscriber_add_item(MicrofeedSubscriber* subscriber, const char* publisher_identifier, const char* uri, MicrofeedItem* item, MicrofeedSubscriberErrorCallback callback, void* user_data) {
 	int retvalue = 0;
 	Publisher* publisher;
@@ -139,7 +164,7 @@ int microfeed_subscriber_add_item(MicrofeedSubscriber* subscriber, const char* p
 	const char* value;
 	
 	if ((publisher = get_publisher(subscriber, publisher_identifier, uri, microfeed_item_get_uid(item), callback, user_data))) {
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_ADD_ITEM);
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_ADD_ITEM);
 		dbus_message_iter_init_append(message, &iter);
 		uid = microfeed_item_get_uid(item);
 		dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &uri);
@@ -170,7 +195,7 @@ int microfeed_subscriber_modify_item(MicrofeedSubscriber*subscriber, const char*
 	const char* value;
 	
 	if ((publisher = get_publisher(subscriber, publisher_identifier, uri, microfeed_item_get_uid(item), callback, user_data))) {
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_MODIFY_ITEM);
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_MODIFY_ITEM);
 		dbus_message_iter_init_append(message, &iter);
 		uid = microfeed_item_get_uid(item);
 		dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &uri);
@@ -196,7 +221,7 @@ int microfeed_subscriber_remove_item(MicrofeedSubscriber* subscriber, const char
 	DBusMessage* message;
 	
 	if ((publisher = get_publisher(subscriber, publisher_identifier, uri, uid, callback, user_data))) {
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_REMOVE_ITEM);
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_REMOVE_ITEM);
 		dbus_message_append_args(message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_STRING, &uid, DBUS_TYPE_INVALID);
 		call_publisher_method(publisher, uri, uid, message, callback, user_data);
 		dbus_message_unref(message);
@@ -220,9 +245,8 @@ int microfeed_subscriber_subscribe_feed(MicrofeedSubscriber* subscriber, const c
 			}
 		} else {
 			feed = feed_new(publisher, uri, callbacks, user_data);
-			microfeed_store_insert(publisher->feeds, feed);
 		
-			message = dbus_message_new_method_call(feed->publisher->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_SUBSCRIBE_FEED);
+			message = dbus_message_new_method_call(feed->publisher->provider->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_SUBSCRIBE_FEED);
 			dbus_message_append_args(message, DBUS_TYPE_STRING, &feed->uri, DBUS_TYPE_INVALID);
 			call_publisher_method(publisher, uri, NULL, message, callback, user_data);
 			dbus_message_unref(message);
@@ -240,15 +264,17 @@ int microfeed_subscriber_unsubscribe_feed(MicrofeedSubscriber* subscriber, const
 	Publisher* publisher;
 	
 	if ((feed = get_feed(subscriber, publisher_identifier, uri, NULL, callback, user_data))) {
-		message = dbus_message_new_method_call(feed->publisher->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_UNSUBSCRIBE_FEED);
+		message = dbus_message_new_method_call(feed->publisher->provider->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_UNSUBSCRIBE_FEED);
 		dbus_message_append_args(message, DBUS_TYPE_STRING, &feed->uri, DBUS_TYPE_INVALID);
 		call_publisher_method(feed->publisher, feed->uri, NULL, message, callback, user_data);
 		dbus_message_unref(message);
 		retvalue = 1;
 	
 		publisher = feed->publisher;
-		microfeed_store_remove(feed->publisher->feeds, feed);
 		feed_free(feed);
+		if (microfeed_store_get_size(publisher->feeds) == 0) {
+			publisher_unref(publisher);
+		}
 	}
 	
 	return retvalue;
@@ -260,7 +286,7 @@ int microfeed_subscriber_update_feed(MicrofeedSubscriber* subscriber, const char
 	DBusMessage* message;
 	
 	if ((feed = get_feed(subscriber, publisher_identifier, uri, NULL, callback, user_data))) {
-		message = dbus_message_new_method_call(feed->publisher->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_UPDATE_FEED);
+		message = dbus_message_new_method_call(feed->publisher->provider->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_UPDATE_FEED);
 		dbus_message_append_args(message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_INVALID);
 		call_publisher_method(feed->publisher, feed->uri, NULL, message, callback, user_data);
 		dbus_message_unref(message);
@@ -279,7 +305,7 @@ int microfeed_subscriber_republish_items(MicrofeedSubscriber* subscriber, const
 	dbus_uint16_t max_count_dbus;
 
 	if ((feed = get_feed(subscriber, publisher_identifier, uri, start_uid, callback, user_data))) {
-		message = dbus_message_new_method_call(feed->publisher->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_REPUBLISH_ITEMS);
+		message = dbus_message_new_method_call(feed->publisher->provider->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_REPUBLISH_ITEMS);
 		start_uid_dbus = (start_uid ? start_uid : "");
 		end_uid_dbus = (end_uid ? end_uid : "");
 		max_count_dbus = max_count;
@@ -298,7 +324,7 @@ int microfeed_subscriber_send_item_data(MicrofeedSubscriber* subscriber, const c
 	DBusMessage* message;
 	
 	if ((publisher = get_publisher(subscriber, publisher_identifier, NULL, uid, callback, user_data))) {
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_SEND_ITEM_DATA);
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_SEND_ITEM_DATA);
 		dbus_message_append_args(message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_STRING, &uid, DBUS_TYPE_INVALID);
 		call_publisher_method(publisher, NULL, uid, message, callback, user_data);
 		dbus_message_unref(message);
@@ -313,22 +339,13 @@ int microfeed_subscriber_create_publisher(MicrofeedSubscriber* subscriber, const
 	Publisher* publisher;
 	DBusMessage* message;
 	
-	if ((publisher = microfeed_store_get(subscriber->publishers_by_identifier, publisher_identifier, Publisher))) {
-		if (callback) {
-			callback(subscriber, publisher_identifier, NULL, NULL, MICROFEED_ERROR_PUBLISHER_ALREADY_EXISTS, "Trying to create a publisher that already exists.", user_data);
-		}
-	} else if (!(publisher = publisher_new(subscriber, publisher_identifier))) {
-		if (callback) {
-			callback(subscriber, publisher_identifier, NULL, NULL, MICROFEED_ERROR_INVALID_PUBLISHER_IDENTIFIER, "Publisher identifier was invalid.", user_data);
-		}
-	} else if (microfeed_configuration_get_publisher_directory(subscriber->configuration, publisher_identifier)) {
+	microfeed_configuration_invalidate(subscriber->configuration);
+	if (microfeed_configuration_get_publisher_directory(subscriber->configuration, publisher_identifier)) {
 		if (callback) {
 			callback(subscriber, publisher_identifier, NULL, NULL, MICROFEED_ERROR_PUBLISHER_ALREADY_EXISTS, "Trying to create a publisher that already exists.", user_data);
 		}		
-	} else {
-		microfeed_store_insert(subscriber->publishers_by_identifier, publisher);
-
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_CREATE_PUBLISHER);
+	} else if ((publisher = get_publisher(subscriber, publisher_identifier, NULL, NULL, callback, user_data))) {
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_CREATE_PUBLISHER);
 		call_publisher_method(publisher, NULL, NULL, message, callback, user_data);
 		dbus_message_unref(message);
 		retvalue = 1;
@@ -343,7 +360,7 @@ int microfeed_subscriber_destroy_publisher(MicrofeedSubscriber* subscriber, cons
 	DBusMessage* message;
 	
 	if ((publisher = get_publisher(subscriber, publisher_identifier, NULL, NULL, callback, user_data))) {
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_DESTROY_PUBLISHER);
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_DESTROY_PUBLISHER);
 		call_publisher_method(publisher, NULL, NULL, message, callback, user_data);
 		dbus_message_unref(message);
 		retvalue = 1;
@@ -358,7 +375,7 @@ int microfeed_subscriber_mark_item(MicrofeedSubscriber* subscriber, const char*
 	DBusMessage* message;
 	
 	if ((publisher = get_publisher(subscriber, publisher_identifier, NULL, NULL, callback, user_data))) {
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_MARK_ITEM);
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_MARK_ITEM);
 		dbus_message_append_args(message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_STRING, &uid, DBUS_TYPE_INVALID);
 		call_publisher_method(publisher, uri, uid, message, callback, user_data);
 		dbus_message_unref(message);
@@ -374,7 +391,7 @@ int microfeed_subscriber_unmark_item(MicrofeedSubscriber* subscriber, const char
 	DBusMessage* message;
 	
 	if ((publisher = get_publisher(subscriber, publisher_identifier, NULL, NULL, callback, user_data))) {
-		message = dbus_message_new_method_call(publisher->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_UNMARK_ITEM);
+		message = dbus_message_new_method_call(publisher->provider->bus_name, publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_UNMARK_ITEM);
 		dbus_message_append_args(message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_STRING, &uid, DBUS_TYPE_INVALID);
 		call_publisher_method(publisher, uri, uid, message, callback, user_data);
 		dbus_message_unref(message);
@@ -392,7 +409,7 @@ int microfeed_subscriber_read_items(MicrofeedSubscriber* subscriber, const char*
 	const char* end_uid_dbus;
 
 	if ((feed = get_feed(subscriber, publisher_identifier, uri, start_uid, callback, user_data))) {
-		message = dbus_message_new_method_call(feed->publisher->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_READ_ITEMS);
+		message = dbus_message_new_method_call(feed->publisher->provider->bus_name, feed->publisher->object_path, MICROFEED_DBUS_INTERFACE_PUBLISHER, MICROFEED_METHOD_NAME_READ_ITEMS);
 		start_uid_dbus = (start_uid ? start_uid : "");
 		end_uid_dbus = (end_uid ? end_uid : "");
 		dbus_message_append_args(message, DBUS_TYPE_STRING, &uri, DBUS_TYPE_STRING, &start_uid_dbus, DBUS_TYPE_STRING, &end_uid_dbus, DBUS_TYPE_INVALID);
@@ -410,61 +427,182 @@ MicrofeedConfiguration* microfeed_subscriber_get_configuration(MicrofeedSubscrib
 	return subscriber->configuration;
 }
 
-
-static Publisher* publisher_new(MicrofeedSubscriber* subscriber, const char* identifier) {
+int microfeed_subscriber_handle_configured_subscriptions(MicrofeedSubscriber* subscriber, MicrofeedSubscriberConfiguredSubscribeCallback configured_subscribe, MicrofeedSubscriberConfiguredUnsubscribeCallback configured_unsubscribe, void* user_data) {
+	int retvalue = 0;
+	const char** subscriptions;
+	const char** s;
+	MicrofeedStoreIterator* publishers_iterator;
 	Publisher* publisher;
-	char* colon;
+	MicrofeedStoreIterator* feeds_iterator;
+	Feed* feed;
+	
+	microfeed_configuration_invalidate(subscriber->configuration);
+	if ((subscriptions = microfeed_configuration_get_subscriptions(subscriber->configuration, subscriber->identifier))) {
+		retvalue = 1;
 
+		/* Subscribe new publishers. */
+		for (s = subscriptions; *s; s++) {
+			if (!microfeed_store_get(subscriber->publishers_by_identifier, *s, Publisher)) {
+				configured_subscribe(subscriber, *s, user_data);
+			}
+		}
+
+		/* Unsubscribe non-existing feeds. */
+		for (publishers_iterator = microfeed_store_iterate(subscriber->publishers_by_identifier, NULL);
+		     (publisher = microfeed_store_iterator_get(publishers_iterator, Publisher));
+		     microfeed_store_iterator_next(publishers_iterator)) {
+			for (s = subscriptions; *s; s++) {
+				if (!strcmp(publisher->identifier, *s)) {
+					break;
+				}		
+			}
+			if (!*s) {
+				for (feeds_iterator = microfeed_store_iterate(publisher->feeds, NULL);
+				     (feed = microfeed_store_iterator_get(feeds_iterator, Feed));
+				     microfeed_store_iterator_next(feeds_iterator)) {
+					configured_unsubscribe(subscriber, publisher->identifier, feed->uri, user_data);
+				}
+			}
+		}
+	}
+
+	return retvalue;
+}
+
+static Provider* provider_new(MicrofeedSubscriber* subscriber, const char* bus_name) {
+	Provider* provider;
+	
+	provider = microfeed_memory_allocate(Provider);
+	provider->subscriber = subscriber;
+	provider->bus_name = strdup(bus_name);
+	provider->publishers_by_object_path = microfeed_store_new_sorted((MicrofeedStoreCompareKeysFunction)strcmp,
+	                                                                 (MicrofeedStoreGetKeyFunction)publisher_get_object_path);
+
+	microfeed_store_insert(subscriber->providers_by_bus_name, provider);
+	
+	return provider;
+}
+
+static void provider_free(Provider* provider) {
+	char buffer[1024];
+
+	microfeed_store_remove(provider->subscriber->providers_by_bus_name, provider);
+	if (provider->unique_connection_name) {
+		microfeed_store_remove(provider->subscriber->providers_by_unique_connection_name, provider);
+	
+		snprintf(buffer, 1024, "type='signal',sender='org.freedesktop.DBus',interface='org.freedesktop.DBus',member='NameOwnerChanged',arg0='%s'", provider->unique_connection_name);
+		dbus_bus_remove_match(provider->subscriber->connection, buffer, NULL);
+		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER);
+		dbus_bus_remove_match(provider->subscriber->connection, buffer, NULL);
+		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',destination='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER_TO_DESTINATION, dbus_bus_get_unique_name(provider->subscriber->connection));
+		dbus_bus_remove_match(provider->subscriber->connection, buffer, NULL);
+		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR);
+		dbus_bus_remove_match(provider->subscriber->connection, buffer, NULL);
+		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',destination='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR_TO_DESTINATION, dbus_bus_get_unique_name(provider->subscriber->connection));
+		dbus_bus_remove_match(provider->subscriber->connection, buffer, NULL);
+	}
+
+	free(provider->bus_name);
+	free(provider->unique_connection_name);
+	microfeed_store_free(provider->publishers_by_object_path);
+	provider->subscriber = NULL;
+	provider->bus_name = NULL;
+	provider->unique_connection_name = NULL;
+	provider->publishers_by_object_path = NULL;
+	microfeed_memory_free(provider);
+}
+
+static const char* provider_get_bus_name(Provider* provider) {
+
+	return provider->bus_name;
+}
+
+static const char* provider_get_unique_connection_name(Provider* provider) {
+
+	return (provider->unique_connection_name ? provider->unique_connection_name : "");
+}
+
+static void provider_set_unique_connection_name(Provider* provider, const char* unique_connection_name) {
+	char buffer[1024];
+
+	provider->unique_connection_name = strdup(unique_connection_name);
+	microfeed_store_insert(provider->subscriber->providers_by_unique_connection_name, provider);
+
+	snprintf(buffer, 1024, "type='signal',sender='org.freedesktop.DBus',interface='org.freedesktop.DBus',member='NameOwnerChanged',arg0='%s'", provider->unique_connection_name);
+	dbus_bus_add_match(provider->subscriber->connection, buffer, NULL);
+	snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER);
+	dbus_bus_add_match(provider->subscriber->connection, buffer, NULL);
+	snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',destination='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER_TO_DESTINATION, dbus_bus_get_unique_name(provider->subscriber->connection));
+	dbus_bus_add_match(provider->subscriber->connection, buffer, NULL);
+	snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR);
+	dbus_bus_add_match(provider->subscriber->connection, buffer, NULL);
+	snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',destination ='%s'", provider->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR_TO_DESTINATION, dbus_bus_get_unique_name(provider->subscriber->connection));
+	dbus_bus_add_match(provider->subscriber->connection, buffer, NULL);
+}
+
+static Publisher* publisher_new(Provider* provider, const char* identifier) {
+	Publisher* publisher = NULL;
+	char* colon;
+	
 	publisher = microfeed_memory_allocate(Publisher);
-	publisher->subscriber = subscriber;
-	publisher->identifier = strdup(identifier); 
-	if (!(colon = strchr(publisher->identifier, ':')) || colon[1] == 0) {
-		free(publisher->identifier);
-		microfeed_memory_free(publisher);
-		publisher = NULL;
-	} else {
+	publisher->reference_count = 1;
+	publisher->identifier = strdup(identifier);
+	if ((colon = strchr(publisher->identifier, ':'))) {
 		*colon = 0;
 		publisher->object_path = microfeed_util_string_concatenate(MICROFEED_DBUS_OBJECT_PATH_PREFIX_PUBLISHER, publisher->identifier, NULL);
-		publisher->bus_name = strdup(colon + 1);
 		*colon = ':';
-		publisher->unique_connection_name = NULL;
-		publisher->last_activity = 0;
-		publisher->feeds = microfeed_store_new_sorted((MicrofeedStoreCompareKeysFunction)strcmp,
-		                                       (MicrofeedStoreGetKeyFunction)feed_get_uri);
+	} else {
+		publisher->object_path = strdup(identifier);
 	}
+	publisher->feeds = microfeed_store_new_sorted((MicrofeedStoreCompareKeysFunction)strcmp,
+		                                      (MicrofeedStoreGetKeyFunction)feed_get_uri);
+	publisher->provider = provider;
+		
+	microfeed_store_insert(provider->publishers_by_object_path, publisher);
+	microfeed_store_insert(provider->subscriber->publishers_by_identifier, publisher);
 	
 	return publisher;
 }
 
 static void publisher_free(Publisher* publisher) {
-	char buffer[1024];
-
-	if (publisher->unique_connection_name) {
-		snprintf(buffer, 1024, "type='signal',sender='org.freedesktop.DBus',interface='org.freedesktop.DBus',member='NameOwnerChanged',arg0='%s'", publisher->unique_connection_name);
-		dbus_bus_remove_match(publisher->subscriber->connection, buffer, NULL);
-		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s'", publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER, publisher->object_path);
-		dbus_bus_remove_match(publisher->subscriber->connection, buffer, NULL);
-		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s',destination='%s'", publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER_TO_DESTINATION, publisher->object_path, dbus_bus_get_unique_name(publisher->subscriber->connection));
-		dbus_bus_remove_match(publisher->subscriber->connection, buffer, NULL);
-		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s'", publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR, publisher->object_path);
-		dbus_bus_remove_match(publisher->subscriber->connection, buffer, NULL);
-		snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s',destination ='%s'", publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR_TO_DESTINATION, publisher->object_path, dbus_bus_get_unique_name(publisher->subscriber->connection));
-		dbus_bus_remove_match(publisher->subscriber->connection, buffer, NULL);
+	microfeed_store_remove(publisher->provider->subscriber->publishers_by_identifier, publisher);
+	microfeed_store_remove(publisher->provider->publishers_by_object_path, publisher);
+	if (microfeed_store_get_size(publisher->provider->publishers_by_object_path) == 0) {
+		provider_free(publisher->provider);
 	}
 	
 	free(publisher->identifier);
 	free(publisher->object_path);
-	free(publisher->bus_name);
-	free(publisher->unique_connection_name);
-	publisher->subscriber = NULL;
+	publisher->provider = NULL;
 	publisher->identifier = NULL;
 	publisher->object_path = NULL;
-	publisher->bus_name = NULL;
-	publisher->unique_connection_name = NULL;
 	microfeed_store_free(publisher->feeds);
 	microfeed_memory_free(publisher);
 }
 
+Publisher* publisher_ref(Publisher* publisher) {
+	publisher->reference_count++;
+	
+	return publisher;
+}
+
+void publisher_unref(Publisher* publisher) {
+	publisher->reference_count--;
+	if (publisher->reference_count == 0) {
+		publisher_free(publisher);
+	}
+}
+
+static const char* publisher_get_identifier(Publisher* publisher) {
+	
+	return publisher->identifier;
+}
+
+static const char* publisher_get_object_path(Publisher* publisher) {
+	
+	return publisher->object_path;
+}
+
 static Feed* feed_new(Publisher* publisher, const char* uri, MicrofeedSubscriberCallbacks* callbacks, void* user_data) {
 	Feed* feed;
 	
@@ -474,11 +612,15 @@ static Feed* feed_new(Publisher* publisher, const char* uri, MicrofeedSubscriber
 	feed->uri = strdup(uri);
 	feed->callbacks = *callbacks;
 	feed->user_data = user_data;
+
+	microfeed_store_insert(publisher->feeds, feed);
 	
 	return feed;
 }
 	
 static void feed_free(Feed* feed) {
+	microfeed_store_remove(feed->publisher->feeds, feed);
+
 	free(feed->uri);
 	free(feed->name);
 	feed->publisher = NULL;
@@ -501,6 +643,12 @@ void feed_unref(Feed* feed) {
 	}
 }
 */
+
+static const char* feed_get_uri(Feed* feed) {
+
+	return feed->uri;
+}
+
 static void signal_feed_update_started(MicrofeedSubscriber* subscriber, Publisher* publisher, DBusMessage* message) {
 	Feed* feed;
 	
@@ -614,38 +762,28 @@ static void signal_item_status_changed(MicrofeedSubscriber* subscriber, Publishe
 	}
 }
 
-static const char* feed_get_uri(Feed* feed) {
-
-	return feed->uri;
-}
-
-static const char* publisher_get_unique_connection_name(Publisher* publisher) {
-	
-	return publisher->unique_connection_name;
-}
-
-static const char* publisher_get_identifier(Publisher* publisher) {
-	
-	return publisher->identifier;
-}
-
 static Publisher* get_publisher(MicrofeedSubscriber* subscriber, const char* publisher_identifier, const char* uri, const char* uid, MicrofeedSubscriberErrorCallback callback, void* user_data) {
-	Publisher* publisher = NULL;
+	Publisher* publisher;
+	const char* bus_name;
+	Provider* provider;
 	
 	if (!(publisher = microfeed_store_get(subscriber->publishers_by_identifier, publisher_identifier, Publisher))) {
-		if (!(publisher = publisher_new(subscriber, publisher_identifier))) {
-			if (callback) {
-				callback(subscriber, publisher_identifier, uri, uid, MICROFEED_ERROR_INVALID_PUBLISHER_IDENTIFIER, "Publisher identifier was invalid.", user_data);
-			}	    
-		} else if (!microfeed_configuration_get_provider_name(subscriber->configuration, publisher->bus_name)) {
-			publisher_free(publisher);
-			publisher = NULL;
-			if (callback) {
-				callback(subscriber, publisher_identifier, uri, uid, MICROFEED_ERROR_NO_SUCH_PROVIDER, "Provider for the publisher does not exist.", user_data);
+		if ((bus_name = strchr(publisher_identifier, ':')) && bus_name[1] != 0) {
+			bus_name++;
+			if (!(provider = microfeed_store_get(subscriber->providers_by_bus_name, bus_name, Provider))) {
+				microfeed_configuration_invalidate(subscriber->configuration);
+				if (microfeed_configuration_get_provider_name(subscriber->configuration, bus_name)) {
+					provider = provider_new(subscriber, bus_name);
+				} else if (callback) {
+					callback(subscriber, publisher_identifier, uri, uid, MICROFEED_ERROR_NO_SUCH_PROVIDER, "Provider for the publisher does not exist.", user_data);
+				}
 			}
-		} else {
-			microfeed_store_insert(subscriber->publishers_by_identifier, publisher);
-		}
+			if (provider) {
+				publisher = publisher_new(provider, publisher_identifier);
+			}
+		} else if (callback) {
+			callback(subscriber, publisher_identifier, uri, uid, MICROFEED_ERROR_INVALID_PUBLISHER_IDENTIFIER, "Publisher identifier was invalid.", user_data);
+		}	    
 	}
 
 	return publisher;
@@ -744,13 +882,12 @@ static void handle_publisher_method_return(DBusPendingCall* pending, void* user_
 	DBusError error;
 	const char* error_name;
 	const char* error_message;
-	char buffer[1024];
 	
 	data = (MethodReturnData*)user_data;
 	reply = dbus_pending_call_steal_reply(pending);
 	if (!reply) {
 		if (data->callback) {
-			data->callback(data->publisher->subscriber, data->publisher->identifier, data->uri, data->uid, MICROFEED_ERROR_DBUS_MESSAGE_FAILED, "No reply from the publisher.", data->user_data);
+			data->callback(data->publisher->provider->subscriber, data->publisher->identifier, data->uri, data->uid, MICROFEED_ERROR_DBUS_MESSAGE_FAILED, "No reply from the publisher.", data->user_data);
 		}
 	} else if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
 		if (data->callback) {
@@ -761,38 +898,21 @@ static void handle_publisher_method_return(DBusPendingCall* pending, void* user_
 			if (!dbus_message_get_args(reply, &error, DBUS_TYPE_STRING, &error_message, DBUS_TYPE_INVALID)) {
 				error_message = NULL;
 			}
-			data->callback(data->publisher->subscriber, data->publisher->identifier, data->uri, data->uid, error_name, error_message, data->user_data);
+			data->callback(data->publisher->provider->subscriber, data->publisher->identifier, data->uri, data->uid, error_name, error_message, data->user_data);
 		}
 	} else {
-		if (!data->publisher->unique_connection_name) {
-			data->publisher->unique_connection_name = strdup(dbus_message_get_sender(reply));
-			microfeed_store_insert(data->publisher->subscriber->publishers_by_unique_connection_name, data->publisher);
-
-			snprintf(buffer, 1024, "type='signal',sender='org.freedesktop.DBus',interface='org.freedesktop.DBus',member='NameOwnerChanged',arg0='%s'", data->publisher->unique_connection_name);
-			dbus_bus_add_match(data->publisher->subscriber->connection, buffer, NULL);
-			snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s'", data->publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER, data->publisher->object_path);
-			dbus_bus_add_match(data->publisher->subscriber->connection, buffer, NULL);
-			snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s',destination='%s'", data->publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_PUBLISHER_TO_DESTINATION, data->publisher->object_path, dbus_bus_get_unique_name(data->publisher->subscriber->connection));
-			dbus_bus_add_match(data->publisher->subscriber->connection, buffer, NULL);
-			snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s'", data->publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR, data->publisher->object_path);
-			dbus_bus_add_match(data->publisher->subscriber->connection, buffer, NULL);
-			snprintf(buffer, 1024, "type='signal',sender='%s',interface='%s',path='%s',destination ='%s'", data->publisher->unique_connection_name, MICROFEED_DBUS_INTERFACE_ERROR_TO_DESTINATION, data->publisher->object_path, dbus_bus_get_unique_name(data->publisher->subscriber->connection));
-			dbus_bus_add_match(data->publisher->subscriber->connection, buffer, NULL);
+		if (!data->publisher->provider->unique_connection_name) {
+			provider_set_unique_connection_name(data->publisher->provider, dbus_message_get_sender(reply));
 		}
 		if (data->callback) {
-			data->callback(data->publisher->subscriber, data->publisher->identifier, data->uri, data->uid, NULL, NULL, data->user_data);
+			data->callback(data->publisher->provider->subscriber, data->publisher->identifier, data->uri, data->uid, NULL, NULL, data->user_data);
 		}
 	}
 	if (reply) {
 		dbus_message_unref(reply);
 	}
-	
-	if (microfeed_store_get_size(data->publisher->feeds) == 0) {
-		microfeed_store_remove(data->publisher->subscriber->publishers_by_identifier, data->publisher);
-		microfeed_store_remove(data->publisher->subscriber->publishers_by_unique_connection_name, data->publisher);
-		publisher_free(data->publisher);
-	}
-	
+		
+	publisher_unref(data->publisher);	
 	free(data->uri);
 	free(data->uid);
 	microfeed_memory_free(data);		
@@ -802,10 +922,10 @@ static void call_publisher_method(Publisher* publisher, const char* uri, const c
 	DBusPendingCall* pending_call;
 	MethodReturnData* data;
 	
-	dbus_connection_send_with_reply(publisher->subscriber->connection, message, &pending_call, -1);
+	dbus_connection_send_with_reply(publisher->provider->subscriber->connection, message, &pending_call, -1);
 	if (pending_call) {
 		data = microfeed_memory_allocate(MethodReturnData);
-		data->publisher = publisher;
+		data->publisher = publisher_ref(publisher);
 		if (uri) {
 			data->uri = strdup(uri);
 		} else {
@@ -831,9 +951,11 @@ static DBusHandlerResult object_message(DBusConnection* connection, DBusMessage*
 	const char* name;
 	const char* old_owner;
 	const char* new_owner;
-	MicrofeedStoreIterator* iterator;
+	MicrofeedStoreIterator* publisher_iterator;
+	MicrofeedStoreIterator* feed_iterator;
 	Feed* feed;
 	int i;
+	Provider* provider;
 	Publisher* publisher;
 	DBusMessage* reply;
 	DBusError error;
@@ -845,19 +967,24 @@ static DBusHandlerResult object_message(DBusConnection* connection, DBusMessage*
 	subscriber = (MicrofeedSubscriber*)user_data;
 	if (dbus_message_is_signal(message, "org.freedesktop.DBus", "NameOwnerChanged") &&
 	    dbus_message_get_args(message, NULL, DBUS_TYPE_STRING, &name, DBUS_TYPE_STRING, &old_owner, DBUS_TYPE_STRING, &new_owner, DBUS_TYPE_INVALID) &&
-	    name[0] == ':' && new_owner[0] == 0 && !strcmp(name, old_owner) && (publisher = microfeed_store_get(subscriber->publishers_by_unique_connection_name, name, Publisher))) {
-		for (iterator = microfeed_store_iterate(publisher->feeds, NULL); (feed = microfeed_store_iterator_get(iterator, Feed)); microfeed_store_iterator_next(iterator)) {
-			feed->callbacks.error_occured(subscriber, publisher->identifier, feed->uri, NULL, MICROFEED_ERROR_PROVIDER_CLOSED_CONNECTION, "Provider closed connection", feed->user_data);	
-			feed_free(feed);
+	    name[0] == ':' && new_owner[0] == 0 && !strcmp(name, old_owner) && (provider = microfeed_store_get(subscriber->providers_by_unique_connection_name, name, Provider))) {
+		for (publisher_iterator = microfeed_store_iterate(provider->publishers_by_object_path, NULL);
+		     (publisher = microfeed_store_iterator_get(publisher_iterator, Publisher));
+		     microfeed_store_iterator_next(publisher_iterator)) {
+			for (feed_iterator = microfeed_store_iterate(publisher->feeds, NULL);
+			     (feed = microfeed_store_iterator_get(feed_iterator, Feed));
+			     microfeed_store_iterator_next(feed_iterator)) {
+				feed->callbacks.error_occured(subscriber, publisher->identifier, feed->uri, NULL, MICROFEED_ERROR_PROVIDER_CLOSED_CONNECTION, "Provider closed connection", feed->user_data);	
+				feed_free(feed);
+			}
+			microfeed_store_iterator_free(feed_iterator);
+			publisher_unref(publisher);
 		}
-/* TODO: If there is a pending call, the return of it will crash because publisher has been destroyed. */
-		microfeed_store_iterator_free(iterator);
-		microfeed_store_remove(subscriber->publishers_by_unique_connection_name, publisher);
-		microfeed_store_remove(subscriber->publishers_by_identifier, publisher);
-		publisher_free(publisher);
+		microfeed_store_iterator_free(publisher_iterator);
 	}
 
-	if ((publisher = microfeed_store_get(subscriber->publishers_by_unique_connection_name, dbus_message_get_sender(message), Publisher))) {
+	if ((provider = microfeed_store_get(subscriber->providers_by_unique_connection_name, dbus_message_get_sender(message), Provider)) &&
+	    (publisher = microfeed_store_get(provider->publishers_by_object_path, dbus_message_get_path(message), Publisher))) {
 		publisher->last_activity = time(NULL);
 		if (dbus_message_is_method_call(message, MICROFEED_DBUS_INTERFACE_SUBSCRIBER, MICROFEED_METHOD_NAME_PING)) {
 			reply = dbus_message_new_method_return(message);
